@@ -1,4 +1,4 @@
-# Copyright (C) 2013-2014 Canonical Ltd.
+# Copyright (C) 2013-2015 Canonical Ltd.
 # Author: Barry Warsaw <barry@ubuntu.com>
 
 # This program is free software: you can redistribute it and/or modify
@@ -20,6 +20,8 @@ __all__ = [
     'MiB',
     'as_loglevel',
     'as_object',
+    'as_port',
+    'as_stripped',
     'as_timedelta',
     'atomic',
     'calculate_signature',
@@ -34,7 +36,6 @@ __all__ = [
 
 import os
 import re
-import time
 import random
 import shutil
 import logging
@@ -46,11 +47,12 @@ from hashlib import sha256
 from importlib import import_module
 
 
+UNIQUE_MACHINE_ID_FILES = ['/var/lib/dbus/machine-id', '/etc/machine-id']
 LAST_UPDATE_FILE = '/userdata/.last_update'
-UNIQUE_MACHINE_ID_FILE = '/var/lib/dbus/machine-id'
 DEFAULT_DIRMODE = 0o02700
 MiB = 1 << 20
 EMPTYSTRING = ''
+NO_PORT = object()
 
 
 def calculate_signature(fp, hash_class=None):
@@ -79,7 +81,7 @@ def safe_remove(path):
     """Like os.remove() but don't complain if the file doesn't exist."""
     try:
         os.remove(path)
-    except (FileNotFoundError, IsADirectoryError):
+    except (FileNotFoundError, IsADirectoryError, PermissionError):
         pass
 
 
@@ -195,13 +197,26 @@ def as_loglevel(value):
         dbus = 'ERROR'
     main_level = getattr(logging, main, None)
     if main_level is None or not isinstance(main_level, int):
-        raise ValueError
+        raise ValueError(value)
     dbus_level = getattr(logging, dbus, None)
     if dbus_level is None or not isinstance(dbus_level, int):
-        raise ValueError
+        raise ValueError(value)
     return main_level, dbus_level
 
 
+def as_port(value):
+    if value.lower() in ('disabled', 'disable'):
+        return NO_PORT
+    result = int(value)
+    if result < 0:
+        raise ValueError(value)
+    return result
+
+
+def as_stripped(value):
+    return value.strip()
+
+
 @contextmanager
 def temporary_directory(*args, **kws):
     """A context manager that creates a temporary directory.
@@ -227,30 +242,24 @@ def makedirs(dir, mode=DEFAULT_DIRMODE):
 def last_update_date():
     """Return the last update date.
 
-    Taken from the mtime of the following files, in order:
-
-    - /userdata/.last_update
-    - /etc/system-image/channel.ini
-    - /etc/ubuntu-build
-
-    First existing path wins.
+    If /userdata/.last_update exists, we use this file's mtime.  If it doesn't
+    exist, then we use the latest mtime of any of the files in
+    /etc/system-image/config.d/*.ini (or whatever directory was given with the
+    -C/--config option).
     """
     # Avoid circular imports.
     from systemimage.config import config
-    channel_ini = os.path.join(
-        os.path.dirname(config.config_file), 'channel.ini')
-    ubuntu_build = config.system.build_file
-    for path in (LAST_UPDATE_FILE, channel_ini, ubuntu_build):
-        try:
-            # Local time, since we can't know the timezone.
-            timestamp = datetime.fromtimestamp(os.stat(path).st_mtime)
-            # Seconds resolution.
-            timestamp = timestamp.replace(microsecond=0)
-            return str(timestamp)
-        except (FileNotFoundError, PermissionError):
-            pass
-    else:
-        return 'Unknown'
+    try:
+        timestamp = datetime.fromtimestamp(os.stat(LAST_UPDATE_FILE).st_mtime)
+    except (FileNotFoundError, PermissionError):
+        # We fall back to the latest mtime of the config.d/*.ini files.
+        timestamps = sorted(
+            datetime.fromtimestamp(path.stat().st_mtime)
+            for path in config.ini_files)
+        if len(timestamps) == 0:
+            return 'Unknown'
+        timestamp = timestamps[-1]
+    return str(timestamp.replace(microsecond=0))
 
 
 def version_detail(details_string=None):
@@ -270,19 +279,20 @@ def version_detail(details_string=None):
     return details
 
 
-_pp_cache = None
-
-def phased_percentage(*, reset=False):
-    global _pp_cache
-    if _pp_cache is None:
-        with open(UNIQUE_MACHINE_ID_FILE, 'rb') as fp:
-            data = fp.read()
-        now = str(time.time()).encode('us-ascii')
-        r = random.Random()
-        r.seed(data + now)
-        _pp_cache = r.randint(0, 100)
-    try:
-        return _pp_cache
-    finally:
-        if reset:
-            _pp_cache = None
+def phased_percentage(channel, target):
+    # Avoid circular imports.
+    from systemimage.config import config
+    if config.phase_override is not None:
+        return config.phase_override
+    for path in UNIQUE_MACHINE_ID_FILES:
+        try:
+            with open(path, 'r', encoding='utf-8') as fp:
+                machine_id = fp.read().strip()
+                break                               # pragma: no branch
+        except FileNotFoundError:
+            pass
+    else:
+        raise RuntimeError('No machine-id file found')
+    r = random.Random()
+    r.seed('{}.{}.{}'.format(channel, target, machine_id))
+    return r.randint(0, 100)
