@@ -1,4 +1,4 @@
-# Copyright (C) 2013-2014 Canonical Ltd.
+# Copyright (C) 2013-2015 Canonical Ltd.
 # Author: Barry Warsaw <barry@ubuntu.com>
 
 # This program is free software: you can redistribute it and/or modify
@@ -23,6 +23,7 @@ __all__ = [
     'data_path',
     'debug',
     'debuggable',
+    'descriptions',
     'find_dbus_process',
     'get_channels',
     'get_index',
@@ -32,14 +33,18 @@ __all__ = [
     'setup_keyring_txz',
     'setup_keyrings',
     'sign',
+    'terminate_service',
     'touch_build',
+    'wait_for_service',
     'write_bytes',
     ]
 
 
 import os
 import ssl
+import dbus
 import json
+import time
 import gnupg
 import psutil
 import shutil
@@ -47,8 +52,8 @@ import inspect
 import tarfile
 import unittest
 
-from contextlib import ExitStack, contextmanager
-from functools import partial, wraps
+from contextlib import ExitStack, contextmanager, suppress
+from functools import partial, partialmethod, wraps
 from http.server import HTTPServer, SimpleHTTPRequestHandler
 from pathlib import Path
 from pkg_resources import resource_filename, resource_string as resource_bytes
@@ -57,7 +62,6 @@ from systemimage.channel import Channels
 from systemimage.config import Configuration, config
 from systemimage.helpers import MiB, atomic, makedirs, temporary_directory
 from systemimage.index import Index
-from systemimage.state import State
 from threading import Thread
 from unittest.mock import patch
 
@@ -78,7 +82,7 @@ def get_channels(filename):
 
 def data_path(filename):
     return os.path.abspath(
-        resource_filename('systemimage.tests.data', filename))
+   resource_filename('systemimage.tests.data', filename))
 
 
 def make_http_server(directory, port, certpem=None, keypem=None):
@@ -117,6 +121,17 @@ def make_http_server(directory, port, certpem=None, keypem=None):
             except ConnectionResetError:
                 super().handle_one_request()
 
+        def do_HEAD(self):
+            # Just tell the client we have the magic file.
+            if self.path == '/user-agent.txt':
+                self.send_response(200)
+                self.end_headers()
+            else:
+                # Canceling a download can cause our internal server to
+                # see various ignorable errors.  No worries.
+                with suppress(BrokenPipeError, ConnectionResetError):
+                    super().do_HEAD()
+
         def do_GET(self):
             # If we requested the magic 'user-agent.txt' file, send back the
             # value of the User-Agent header.  Otherwise, vend as normal.
@@ -127,12 +142,10 @@ def make_http_server(directory, port, certpem=None, keypem=None):
                 self.end_headers()
                 self.wfile.write(user_agent.encode('utf-8'))
             else:
-                try:
+                # Canceling a download can cause our internal server to
+                # see various ignorable errors.  No worries.
+                with suppress(BrokenPipeError, ConnectionResetError):
                     super().do_GET()
-                except (BrokenPipeError, ConnectionResetError):
-                    # Canceling a download can cause our internal server to
-                    # see various ignorable errors.  No worries.
-                    pass
     # Create the server in the main thread, but start it in the sub-thread.
     # This lets the main thread call .shutdown() to stop everything.  Return
     # just the shutdown method to the caller.
@@ -198,47 +211,98 @@ def make_http_server(directory, port, certpem=None, keypem=None):
         return resources.pop_all()
 
 
-def configuration(function):
-    """Decorator that produces a temporary configuration for testing.
-
-    The config_00.ini template is copied to a temporary file and the the
-    various file system locations are filled in with the location for a,
-    er, temporary temporary directory.  This temporary configuration
-    file is loaded up and the global configuration object is patched so
-    that all other code will see it instead of the default global
-    configuration object.
-
-    Everything is properly cleaned up after the test method exits.
-    """
-    @wraps(function)
-    def wrapper(*args, **kws):
-        with ExitStack() as resources:
-            etc_dir = resources.enter_context(temporary_directory())
-            ini_file = os.path.join(etc_dir, 'client.ini')
-            temp_tmpdir = resources.enter_context(temporary_directory())
-            temp_vardir = resources.enter_context(temporary_directory())
+# This defines the @configuration decorator used in various test suites to
+# create a temporary config.d/ directory for a test.  This is all fairly
+# complicated, but here's what's going on.
+#
+# The _wrapper() function is the inner part of the decorator, and it does the
+# heart of the operation, which is to create a temporary directory for
+# config.d, along with temporary var and tmp directories.  These latter two
+# will be interpolated into any configuration file copied into config.d.
+#
+# The outer decorator function differs depending on whether @configuration was
+# given without arguments, or called with arguments at the time of the
+# function definition.
+#
+# In the former case, e.g.
+#
+# @configuration
+# def test_something(self):
+#
+# The default 00.ini file is interpolated and copied into config.d.  Simple.
+#
+# In the latter case, e.g.
+#
+# @configuration('some-config.ini')
+# def test_something(self):
+#
+# There's actually another level of interior function, because the outer
+# decorator itself is getting called.  Here, any named configuration file is
+# additionally copied to the config.d directory, renaming it sequentionally to
+# something like 01_override.ini, with the numeric part incrementing
+# monotonically.
+#
+# The implementation is tricky because we want the call sites to be simple.
+def _wrapper(self, function, ini_files, *args, **kws):
+    start = 0
+    with ExitStack() as resources:
+        # Create the config.d directory and copy all the source ini files to
+        # this directory in sequential order, interpolating in the temporary
+        # tmp and var directories.
+        config_d = resources.enter_context(temporary_directory())
+        temp_tmpdir = resources.enter_context(temporary_directory())
+        temp_vardir = resources.enter_context(temporary_directory())
+        for ini_file in ini_files:
+            dst = os.path.join(config_d, '{:02d}_override.ini'.format(start))
+            start += 1
             template = resource_bytes(
-                'systemimage.tests.data', 'config_00.ini').decode('utf-8')
-            with atomic(ini_file) as fp:
+                'systemimage.tests.data', ini_file).decode('utf-8')
+            with atomic(dst) as fp:
                 print(template.format(tmpdir=temp_tmpdir,
                                       vardir=temp_vardir), file=fp)
-            config = Configuration(ini_file)
-            resources.enter_context(
-                patch('systemimage.config._config', config))
-            resources.enter_context(
-                patch('systemimage.device.check_output',
-                      return_value='nexus7'))
-            # Make sure the cache_partition and data_partition exist.
-            makedirs(config.updater.cache_partition)
-            makedirs(config.updater.data_partition)
-            # The method under test is allowed to specify some additional
-            # keyword arguments, in order to pass some variables in from the
-            # wrapper.
-            signature = inspect.signature(function)
-            if 'ini_file' in signature.parameters:
-                kws['ini_file'] = ini_file
-            return function(*args, **kws)
-    return wrapper
+        # Patch the global configuration object so that it can be used
+        # directly, which is good enough in most cases.  Also patch the bit of
+        # code that detects the device name.
+        config = Configuration(config_d)
+        resources.enter_context(
+            patch('systemimage.config._config', config))
+        resources.enter_context(
+            patch('systemimage.device.check_output',
+                  return_value='nexus7'))
+        # Make sure the cache_partition and data_partition exist.
+        makedirs(config.updater.cache_partition)
+        makedirs(config.updater.data_partition)
+        # The method under test is allowed to specify some additional
+        # keyword arguments, in order to pass some variables in from the
+        # wrapper.
+        signature = inspect.signature(function)
+        if 'config_d' in signature.parameters:
+            kws['config_d'] = config_d
+        if 'config' in signature.parameters:
+            kws['config'] = config
+        # Call the function with the given arguments and return the result.
+        return function(self, *args, **kws)
+
+
+def configuration(*args):
+    """Outer decorator which can be called or not at function definition time.
+
+    If called, the arguments are positional only, and name the test data .ini
+    files which are to be copied to config.d directory.  If none are given,
+    then 00.ini is used.
+    """
+    if len(args) == 1 and callable(args[0]):
+        # We assume this was the bare @configuration decorator flavor.
+        function = args[0]
+        inner = partialmethod(_wrapper, function, ('00.ini',))
+        return wraps(function)(inner)
+    else:
+        # We assume this was the called @configuration(...) decorator flavor,
+        # so create the actual decorator that wraps the _wrapper function.
+        def decorator(function):
+            inner = partialmethod(_wrapper, function, args)
+            return wraps(function)(inner)
+        return decorator
 
 
 def sign(filename, pubkey_ring):
@@ -249,6 +313,8 @@ def sign(filename, pubkey_ring):
         with.  This keyring must contain only one key, and its key id must
         exist in the master secret keyring.
     """
+    # filename could be a Path object.  For now, just str-ify it.
+    filename = str(filename)
     with ExitStack() as resources:
         home = resources.enter_context(temporary_directory())
         secring = data_path('master-secring.gpg')
@@ -268,7 +334,7 @@ def sign(filename, pubkey_ring):
 
 def copy(filename, todir, dst=None):
     src = data_path(filename)
-    dst = os.path.join(todir, filename if dst is None else dst)
+    dst = os.path.join(str(todir), filename if dst is None else dst)
     makedirs(os.path.dirname(dst))
     shutil.copy(src, dst)
 
@@ -395,21 +461,24 @@ def chmod(path, new_mode):
         os.chmod(path, old_mode)
 
 
-def touch_build(version, timestamp=None):
+def touch_build(version, timestamp=None, use_config=None):
     # LP: #1220238 - assert that no old-style version numbers are being used.
     assert 0 <= version < (1 << 16), (
-        'old style version number: {}'.format(version))
-    with open(config.system.build_file, 'w', encoding='utf-8') as fp:
-        print(version, file=fp)
+        'Old style version number: {}'.format(version))
+    if use_config is None:
+        use_config = config
+    override = Path(use_config.config_d) / '99_build.ini'
+    with override.open('wt', encoding='utf-8') as fp:
+        print("""\
+[service]
+build_number: {}
+""".format(version), file=fp)
+    # We have to touch the mtimes for all the files in the config directory.
     if timestamp is not None:
         timestamp = int(timestamp)
-        os.utime(config.system.build_file, (timestamp, timestamp))
-        channel_ini = os.path.join(
-            os.path.dirname(config.config_file), 'channel.ini')
-        try:
-            os.utime(channel_ini, (timestamp, timestamp))
-        except FileNotFoundError:
-            pass
+        for path in Path(use_config.config_d).iterdir():
+            os.utime(str(path), (timestamp, timestamp))
+    use_config.reload()
 
 
 def write_bytes(path, size_in_mib):
@@ -431,13 +500,13 @@ def debuggable(fn):
 
 
 @contextmanager
-def debug(*, check_flag=False):
+def debug(*, check_flag=False, end='\n'):
     if not check_flag or os.path.exists('/tmp/debug.enabled'):
         path = Path('/tmp/debug.log')
     else:
         path = Path(os.devnull)
     with path.open('a', encoding='utf-8') as fp:
-        function = partial(print, file=fp)
+        function = partial(print, file=fp, end=end)
         function.fp = fp
         yield function
         fp.flush()
@@ -487,6 +556,8 @@ class ServerTestBase(unittest.TestCase):
         SystemImagePlugin.controller.set_mode(cert_pem='cert.pem')
 
     def setUp(self):
+        # Avoid circular imports.
+        from systemimage.state import State
         self._resources = ExitStack()
         self._state = State()
         try:
@@ -544,3 +615,44 @@ class ServerTestBase(unittest.TestCase):
                 dict(type='device-signing'),
                 os.path.join(self._serverdir, self.CHANNEL, self.DEVICE,
                              'device-signing.tar.xz'))
+
+
+def descriptions(path):
+    descriptions = []
+    for image in path:
+        # There's only one description per image so order doesn't
+        # matter.
+        descriptions.extend(image.descriptions.values())
+    return descriptions
+
+
+def wait_for_service(*, restart=False, reload=True):
+    bus = dbus.SystemBus()
+    if restart:
+        service = bus.get_object('com.canonical.SystemImage', '/Service')
+        iface = dbus.Interface(service, 'com.canonical.SystemImage')
+        iface.Exit()
+    service = dbus.SystemBus().get_object('org.freedesktop.DBus', '/')
+    iface = dbus.Interface(service, 'org.freedesktop.DBus')
+    if reload:
+        iface.ReloadConfig()
+    # Wait until the system-image-dbus process is actually running.
+    # http://people.freedesktop.org/~david/eggdbus-20091014/eggdbus-interface-org.freedesktop.DBus.html#eggdbus-method-org.freedesktop.DBus.StartServiceByName
+    reply = 0
+    # 2015-03-09 BAW: This could potentially spin forever, but we'll assume
+    # D-Bus eventually is successful in starting the service.
+    while reply != 2:
+        reply = iface.StartServiceByName('com.canonical.SystemImage', 0)
+        time.sleep(0.1)
+
+
+def terminate_service():
+    # Avoid circular imports.
+    from systemimage.testing.nose import SystemImagePlugin
+    proc = find_dbus_process(SystemImagePlugin.controller.ini_path)
+    if proc is not None:
+        bus = dbus.SystemBus()
+        service = bus.get_object('com.canonical.SystemImage', '/Service')
+        iface = dbus.Interface(service, 'com.canonical.SystemImage')
+        iface.Exit()
+        proc.wait()
